/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;

/**
 * An abstract class for implementing input stream consumers.
 * <p>
 * Provides a {@link #call()}, {@link #cancel()}, and
 * {@link #postProcessData()} implementation.
 */
public abstract class AbstractInputStreamConsumer implements InputStreamConsumer {
  /** The stream to read. */
  protected final InputStream stream;

  /**
   * Creates an instance with the specified stream.
   *
   * @param stream stream from which to read
   *
   * @throws IllegalArgumentException if {@code stream} is {@code null}
   *
   * @since 2.0
   */
  public AbstractInputStreamConsumer(InputStream stream) {
    super();
    if (stream == null) {
      throw new IllegalArgumentException("stream must not be null");
    }
    this.stream = stream;
  }

  /**
   * Calls {@link #read()} until it returns -1, then calls
   * {@link #postProcessData()}, and then calls {@link #data()}, returning what
   * {@link #data()} returns.
   *
   * @since 2.0
   */
  @Override
  public ByteBuffer call() throws IOException {
    for (;;) {
      if (read() != -1) continue;
      postProcessData();
      break;
    }
    return data();
  }

  /**
   * Stops this consumer by closing the input stream.
   *
   * @since 2.0
   */
  @Override
  public void cancel() {
    try {
      stream.close();
    } catch (IOException e) {
      /* IGNORED */
    }
  }

  /**
   * Returns the data consumed from the input stream.
   *
   * @return consumed data
   */
  protected abstract ByteBuffer data();

  /**
   * Reads a chunk of data from the input stream. May block.
   *
   * @return number of bytes read or -1 indicating end-of-stream
   *
   * @throws IOException if an I/O error occurs
   */
  protected abstract int read() throws IOException;

  /**
   * Performs post-processing of data.
   * <p>
   * This method does nothing, but subclasses may override it to do something.
   */
  protected void postProcessData() {
  }
}
