/*
 * Copyright (c) 2014 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.FutureTask;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.RunnableFuture;
import java.util.concurrent.ThreadFactory;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * A {@link ThreadPoolExecutor} that returns
 * {@link java.util.concurrent.Future Future} instances that, when canceled,
 * will invoke {@link CancelableCallable#cancel()} if the submitted
 * {@link Callable} is a {@link CancelableCallable}.
 *
 * @since 2.0
 */
public class CancelableExecutor extends ThreadPoolExecutor {
  /**
   * Creates a fixed size thread pool with the specified number of threads and
   * a shared unbounded queue.
   *
   * @param nThreads number of threads in the pool
   */
  public CancelableExecutor(int nThreads) {
    this(nThreads, nThreads, 0L, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
  }

  /**
   * Creates an instance by invoking the super constructor with the specified
   * arguments.
   *
   * @param corePoolSize same as for super constructor
   * @param maximumPoolSize same as for super constructor
   * @param keepAliveTime same as for super constructor
   * @param unit same as for super constructor
   * @param workQueue same as for super constructor
   *
   * @see ThreadPoolExecutor#ThreadPoolExecutor(int, int, long, TimeUnit,
   *      BlockingQueue)
   */
  public CancelableExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime,
      TimeUnit unit, BlockingQueue<Runnable> workQueue) {
    super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue);
  }

  /**
   * Creates an instance by invoking the super constructor with the specified
   * arguments.
   *
   * @param corePoolSize same as for super constructor
   * @param maximumPoolSize same as for super constructor
   * @param keepAliveTime same as for super constructor
   * @param unit same as for super constructor
   * @param workQueue same as for super constructor
   * @param threadFactory same as for super constructor
   *
   * @see ThreadPoolExecutor#ThreadPoolExecutor(int, int, long, TimeUnit,
   *      BlockingQueue, ThreadFactory)
   */
  public CancelableExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime,
      TimeUnit unit, BlockingQueue<Runnable> workQueue, ThreadFactory threadFactory) {
    super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, threadFactory);
  }

  /**
   * Creates an instance by invoking the super constructor with the specified
   * arguments.
   *
   * @param corePoolSize same as for super constructor
   * @param maximumPoolSize same as for super constructor
   * @param keepAliveTime same as for super constructor
   * @param unit same as for super constructor
   * @param workQueue same as for super constructor
   * @param handler same as for super constructor
   *
   * @see ThreadPoolExecutor#ThreadPoolExecutor(int, int, long, TimeUnit,
   *      BlockingQueue, RejectedExecutionHandler)
   */
  public CancelableExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime,
      TimeUnit unit, BlockingQueue<Runnable> workQueue, RejectedExecutionHandler handler) {
    super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, handler);
  }

  /**
   * Creates an instance by invoking the super constructor with the specified
   * arguments.
   *
   * @param corePoolSize same as for super constructor
   * @param maximumPoolSize same as for super constructor
   * @param keepAliveTime same as for super constructor
   * @param unit same as for super constructor
   * @param workQueue same as for super constructor
   * @param threadFactory same as for super constructor
   * @param handler same as for super constructor
   *
   * @see ThreadPoolExecutor#ThreadPoolExecutor(int, int, long, TimeUnit,
   *      BlockingQueue, ThreadFactory, RejectedExecutionHandler)
   */
  public CancelableExecutor(int corePoolSize, int maximumPoolSize, long keepAliveTime,
      TimeUnit unit, BlockingQueue<Runnable> workQueue, ThreadFactory threadFactory,
      RejectedExecutionHandler handler) {
    super(corePoolSize, maximumPoolSize, keepAliveTime, unit, workQueue, threadFactory, handler);
  }

  @Override
  protected <T> RunnableFuture<T> newTaskFor(final Callable<T> callable) {
    if (callable instanceof CancelableCallable) {
      return new FutureTask<T>(callable) {
        @Override
        public boolean cancel(boolean mayInterruptIfRunning) {
          boolean result;
          try {
            ((CancelableCallable<T>)callable).cancel();
          } finally {
            result = super.cancel(mayInterruptIfRunning);
          }
          return result;
        }
      };
    } else {
      return super.newTaskFor(callable);
    }
  }
}
