/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;

/**
 * Reads an {@link InputStream}, discarding what is read as it goes, until it
 * encounters the end of the stream or an exception is thrown.
 * <p>
 */
public class DiscardingInputStreamConsumer extends AbstractInputStreamConsumer {
  private final int bufferSize;
  private byte[] buffer;

  /**
   * Creates an instance with the specified stream and buffer size.
   *
   * @param stream stream from which to read
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by the input stream
   *
   * @since 2.0
   */
  public DiscardingInputStreamConsumer(InputStream stream, int bufferSize) {
    super(stream);
    this.bufferSize = bufferSize;
    this.buffer = null;
  }

  @Override
  protected int read() throws IOException {
    if (buffer == null) {
      try {
        int numBytesSkipped = (int)stream.skip(Integer.MAX_VALUE);
        if (numBytesSkipped > 0) return numBytesSkipped;
        /*
         * The documentation for the skip method only says the number of
         * skipped bytes is returned. It says nothing about returning -1 to
         * indicate the end of the stream has been reached, so perform an
         * additional read to determine whether we are at the end of the
         * stream.
         */
        int b = stream.read();
        return (b == -1) ? -1 : 1;
      } catch (IOException e) {
        /*
         * Assume skip is not supported by the stream. Switch to using reads
         * into a buffer instead.
         */
        buffer = new byte[bufferSize];
      }
    }
    return stream.read(buffer);
  }

  @Override
  protected ByteBuffer data() {
    return ByteBuffer.allocate(0);
  }

  /**
   * Creates a factory that returns a new instance of this class with the
   * specified buffer size.
   *
   * @param bufferSize size of read buffer used if {@code skip(long)} is not
   *          supported by the input stream
   *
   * @since 2.0
   */
  public static InputStreamConsumerFactory newFactory(final int bufferSize) {
    return new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new DiscardingInputStreamConsumer(s, bufferSize);
      }
    };
  }
}
