/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import java.io.IOException;
import java.io.InputStream;

/**
 * Reads an {@link InputStream}, placing what is read into a fixed-size buffer,
 * until it encounters the end of the stream or an exception is thrown. If it
 * reads more bytes than will fit in the buffer, the buffer will contain the
 * most recent <i>buffer-size</i> bytes read.
 */
public class FixedSizeBufferInputStreamConsumer extends BufferInputStreamConsumer {
  private boolean bufferWrapped;

  /**
   * Creates an instance with the specified stream and buffer size.
   *
   * @param stream stream from which to read
   * @param bufferSize size of the buffer into which data read from the input
   *          stream should be placed
   *
   * @since 2.0
   */
  public FixedSizeBufferInputStreamConsumer(InputStream stream, int bufferSize) {
    super(stream, bufferSize);
    bufferWrapped = false;
  }

  @Override
  protected int read() throws IOException {
    int numBytesRead = stream.read(buffer, bufferPosition, buffer.length - bufferPosition);
    if (numBytesRead != -1) {
      bufferPosition += numBytesRead;
      if (bufferPosition == buffer.length) {
        bufferWrapped = true;
        bufferPosition = 0;
      }
    }
    return numBytesRead;
  }

  @Override
  protected void postProcessData() {
    if (!bufferWrapped) return;
    if (bufferPosition == buffer.length) return;
    if (bufferPosition == 0) {
      bufferPosition = buffer.length;
      return;
    }
    if (bufferPosition < (buffer.length / 2)) {
      byte[] tmp = new byte[bufferPosition];
      System.arraycopy(buffer, 0, tmp, 0, tmp.length);
      System.arraycopy(buffer, bufferPosition, buffer, 0, buffer.length - bufferPosition);
      System.arraycopy(tmp, 0, buffer, buffer.length - bufferPosition, tmp.length);
    } else {
      byte[] tmp = new byte[buffer.length - bufferPosition];
      System.arraycopy(buffer, bufferPosition, tmp, 0, tmp.length);
      System.arraycopy(buffer, 0, buffer, tmp.length, bufferPosition);
      System.arraycopy(tmp, 0, buffer, 0, tmp.length);
    }
    bufferPosition = buffer.length;
  }

  /**
   * Creates a factory that returns a new instance of this class with the
   * specified buffer size.
   *
   * @param bufferSize size of the buffer into which data read from the input
   *          stream should be placed
   *
   * @since 2.0
   */
  public static InputStreamConsumerFactory newFactory(final int bufferSize) {
    return new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new FixedSizeBufferInputStreamConsumer(s, bufferSize);
      }
    };
  }
}
