/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;

/**
 * Reads an {@link InputStream}, placing what is read into an automatically
 * growing buffer, until it encounters the end of the stream or an exception is
 * thrown.
 */
public class GrowingBufferInputStreamConsumer extends BufferInputStreamConsumer {
  /**
   * Default initial buffer size in bytes. Its value is
   * {@value #DEFAULT_BUFFER_SIZE}.
   */
  public static final int DEFAULT_BUFFER_SIZE = 32;

  /**
   * Creates an instance with the specified stream and with an initial buffer
   * size of {@value #DEFAULT_BUFFER_SIZE} bytes. Behaves just like calling the
   * {@link #GrowingBufferInputStreamConsumer(InputStream, int)} constructor
   * with the specified stream and a buffer size of
   * <code>{@value #DEFAULT_BUFFER_SIZE}</code>.
   *
   * @param stream stream from which to read
   *
   * @since 2.0
   */
  public GrowingBufferInputStreamConsumer(InputStream stream) {
    this(stream, DEFAULT_BUFFER_SIZE);
  }

  /**
   * Creates an instance with the specified stream and buffer size.
   *
   * @param stream stream from which to read
   * @param bufferSize initial size of the buffer into which data read from the
   *          input stream should be placed
   *
   * @since 2.0
   */
  public GrowingBufferInputStreamConsumer(InputStream stream, int bufferSize) {
    super(stream, bufferSize);
  }

  @Override
  protected int read() throws IOException {
    int numBytesRead = stream.read(buffer, bufferPosition, buffer.length - bufferPosition);
    if (numBytesRead != -1) {
      bufferPosition += numBytesRead;
      if (bufferPosition == buffer.length) {
        buffer = Arrays.copyOf(buffer, buffer.length * 2);
      }
    }
    return numBytesRead;
  }

  /**
   * Creates a factory that returns a new instance of this class with the
   * default initial buffer size.
   *
   * @since 2.0
   */
  public static InputStreamConsumerFactory newFactory() {
    return new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new GrowingBufferInputStreamConsumer(s);
      }
    };
  }

  /**
   * Creates a factory that returns a new instance of this class with the
   * specified initial buffer size.
   *
   * @param bufferSize initial size of the buffer into which data read from the
   *          input stream should be placed
   *
   * @since 2.0
   */
  public static InputStreamConsumerFactory newFactory(final int bufferSize) {
    return new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new GrowingBufferInputStreamConsumer(s, bufferSize);
      }
    };
  }
}
