/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import java.nio.ByteBuffer;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletionService;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorCompletionService;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Runs an operating system process and consumes data written to its standard
 * output and standard error streams.
 * <p>
 * Interrupting the thread in which a {@link ProcessRunner} is running will
 * cause the {@link Process#destroy() destroy()} method to be called on the
 * {@link Process} object of the operating system process. On a UNIX-like
 * operating system, this will likely result in the process being sent the
 * <tt>SIGTERM</tt> signal. (Unfortunately, the Java 7 API documentation does
 * not guarantee this behavior.) The process is expected to be responsive to
 * the <tt>SIGTERM</tt> signal and exit.
 * <p>
 * If the process creates child processes, it is up to the process to handle
 * what should happen to the child processes if it is destroyed via
 * {@link Process#destroy()}. For example, if the process is <tt>sh</tt>
 * invoked with two arguments, <tt>-c</tt> and <tt>sleep 5</tt>, depending on
 * the implementation of the <tt>sh</tt> shell, the <tt>sleep 5</tt> command
 * may result in a child process. If the <tt>sh</tt> process is destroyed and
 * the <tt>sleep 5</tt> command resulted in a child process, the
 * <tt>sleep 5</tt> child process may continue to run.
 * <p>
 * Better would be to invoke <tt>sh</tt> with the following two arguments:
 * <tt>-c</tt> and <tt>exec sleep 5</tt>. The <tt>exec</tt> causes
 * <tt>sleep 5</tt> to replace the <tt>sh</tt> process so that there is only
 * one process. However, in scenarios where the process needs to run with child
 * processes, and they should die when the process is destroyed, the process
 * will need to arrange for that to happen. For example, it might send each
 * child process a <tt>SIGTERM</tt> signal.
 * <p>
 * <strong>Examples</strong>
 *
 * <p>
 * Here's how to run the <tt>uptime</tt> program by creating a
 * {@code ProcessRunner} instance:
 *
 * <pre> {@code
 * ExecutorService service = null;
 * try {
 *   service = Executors.newSingleThreadExecutor();
 *   ProcessBuilder builder = new ProcessBuilder("uptime");
 *   ProcessRunner runner = new ProcessRunner(builder);
 *   Future<ProcessRunnerResult> future = service.submit(runner);
 *   ProcessRunnerResult result = future.get();
 * } finally {
 *   if (service != null) service.shutdownNow();
 * }}</pre>
 *
 * <p>
 * Here's how to run the same <tt>uptime</tt> program using one of the static
 * convenience methods:
 *
 * <pre> {@code
 * ProcessBuilder builder = new ProcessBuilder("uptime");
 * ProcessRunnerResult result = ProcessRunner.run(builder);}</pre>
 *
 * <p>
 * Here's how to run the same <tt>uptime</tt> program reading its standard
 * output into an 80 byte fixed-size buffer:
 *
 * <pre> {@code
 * ProcessBuilder builder = new ProcessBuilder("uptime");
 * InputStreamConsumerFactory factory;
 * factory = FixedSizeBufferInputStreamConsumer.newFactory(80);
 * ProcessRunnerResult result = ProcessRunner.run(builder, factory);}</pre>
 */
public class ProcessRunner implements Callable<ProcessRunnerResult> {
  private final ProcessBuilder builder;
  private final InputStreamConsumerFactory stdoutConsumerFactory;
  private final InputStreamConsumerFactory stderrConsumerFactory;

  /**
   * Creates an instance with the specified builder and the default
   * standard-output and standard-error consumer factories that create
   * {@link GrowingBufferInputStreamConsumer} instances.
   *
   * @param builder used to start the process when this class runs
   */
  public ProcessRunner(ProcessBuilder builder) {
    this(builder, null);
  }

  /**
   * Creates an instance with the specified builder and standard-output
   * consumer factory and the default standard-error consumer factory that
   * creates {@link GrowingBufferInputStreamConsumer} instances.
   *
   * @param builder used to start the process when this class runs
   * @param stdoutConsumerFactory factory used to create the standard-output
   *          consumer or {@code null} to use the default factory
   *
   * @since 2.0
   */
  public ProcessRunner(ProcessBuilder builder, InputStreamConsumerFactory stdoutConsumerFactory) {
    this(builder, stdoutConsumerFactory, null);
  }

  /**
   * Creates an instance with the specified builder, standard-output consumer
   * factory, and standard-error consumer factory.
   *
   * @param builder used to start the process when this class runs
   * @param stdoutConsumerFactory factory used to create the standard-output
   *          consumer or {@code null} to use the default factory
   * @param stderrConsumerFactory factory used to create the standard-error
   *          consumer or {@code null} to use the default factory
   *
   * @since 2.0
   */
  public ProcessRunner(ProcessBuilder builder, InputStreamConsumerFactory stdoutConsumerFactory,
      InputStreamConsumerFactory stderrConsumerFactory) {
    super();
    this.builder = builder;
    this.stdoutConsumerFactory =
        (stdoutConsumerFactory == null) ? newDefaultConsumerFactory() : stdoutConsumerFactory;
    this.stderrConsumerFactory =
        (stderrConsumerFactory == null) ? newDefaultConsumerFactory() : stderrConsumerFactory;
  }

  /**
   * Runs the process.
   *
   * @return the result of running the process
   *
   * @since 2.0
   */
  @Override
  public ProcessRunnerResult call() {
    ExecutorService pool = null;
    CompletionService<ByteBuffer> completionPool = null;
    Process process = null;
    Future<ByteBuffer> stdoutFuture = null;
    Future<ByteBuffer> stderrFuture = null;
    ByteBuffer stdout = null;
    ByteBuffer stderr = null;
    Integer exitStatus = null;
    Throwable exception = null;

    try {
      pool = new CancelableExecutor(2);
      completionPool = new ExecutorCompletionService<ByteBuffer>(pool);
      process = builder.start();
      stdoutFuture =
          completionPool.submit(stdoutConsumerFactory.newConsumer(process.getInputStream()));
      stderrFuture =
          completionPool.submit(stderrConsumerFactory.newConsumer(process.getErrorStream()));
      completionPool.take().get();
      completionPool.take().get();
      stdout = stdoutFuture.get();
      stderr = stderrFuture.get();
      exitStatus = process.waitFor();
      pool.shutdown();
      pool.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
    } catch (InterruptedException e) {
      exception = e;
      Thread.currentThread().interrupt();
    } catch (ExecutionException e) {
      exception = e.getCause();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      exception = t;
    } finally {
      destroySilently(process);
      waitForSilently(process);
      cancelSilently(stdoutFuture);
      cancelSilently(stderrFuture);
      shutdownNowSilently(pool);
      awaitTerminationSilently(pool);
    }

    return new ProcessRunnerResult(exitStatus, exception, stdout, stderr);
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder, and
   * with the default standard-output consumer and standard-error consumer.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   */
  public static ProcessRunnerResult run(ProcessBuilder builder) throws InterruptedException {
    return run(builder, null);
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder and
   * standard-output consumer, and with the default standard-error consumer.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   * @param stdoutConsumerFactory factory used to create the standard-output
   *          consumer or {@code null} to use the default factory
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   *
   * @since 2.0
   */
  public static ProcessRunnerResult run(ProcessBuilder builder,
      InputStreamConsumerFactory stdoutConsumerFactory) throws InterruptedException {
    return run(builder, stdoutConsumerFactory, null);
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder,
   * standard-output consumer, and standard-error consumer.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   * @param stdoutConsumerFactory factory used to create the standard-output
   *          consumer or {@code null} to use the default factory
   * @param stderrConsumerFactory factory used to create the standard-error
   *          consumer or {@code null} to use the default factory
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   *
   * @since 2.0
   */
  public static ProcessRunnerResult run(ProcessBuilder builder,
      InputStreamConsumerFactory stdoutConsumerFactory,
      InputStreamConsumerFactory stderrConsumerFactory) throws InterruptedException {
    try {
      return run(builder, stdoutConsumerFactory, stderrConsumerFactory, 0L, TimeUnit.MILLISECONDS);
    } catch (TimeoutException e) {
      /* NOT REACHED: Used 0 wait-forever timeout. */
      throw new RuntimeException(e);
    }
  }

  /**
   * Creates and runs a {@code ProcessRunner} with the specified builder,
   * standard-output consumer, standard-error consumer, and timeout.
   *
   * @param builder used by the {@code ProcessRunner} instance to start the
   *          process
   * @param stdoutConsumerFactory factory used to create the standard-output
   *          consumer or {@code null} to use the default factory
   * @param stderrConsumerFactory factory used to create the standard-error
   *          consumer or {@code null} to use the default factory
   * @param timeout amount of time to wait for {@code ProcessRunner} to
   *          complete; 0 means wait forever
   * @param unit unit of {@code timeout} parameter
   *
   * @return result of running {@code ProcessRunner} instance
   *
   * @throws InterruptedException if the current thread is interrupted while
   *           waiting for the {@code ProcessRunner} to finish
   * @throws TimeoutException if the {@code ProcessRunner} has not finished
   *           running within the specified timeout
   *
   * @since 2.0
   */
  public static ProcessRunnerResult run(ProcessBuilder builder,
      InputStreamConsumerFactory stdoutConsumerFactory,
      InputStreamConsumerFactory stderrConsumerFactory, long timeout, TimeUnit unit)
      throws InterruptedException, TimeoutException {
    ExecutorService s = null;
    try {
      ProcessRunner pr = new ProcessRunner(builder, stdoutConsumerFactory, stderrConsumerFactory);
      s = Executors.newSingleThreadExecutor();
      Future<ProcessRunnerResult> f = s.submit(pr);
      return (timeout == 0L) ? f.get() : f.get(timeout, unit);
    } catch (ExecutionException e) {
      return new ProcessRunnerResult(null, e.getCause(), null, null);
    } finally {
      shutdownNowSilently(s);
    }
  }

  private static InputStreamConsumerFactory newDefaultConsumerFactory() {
    return GrowingBufferInputStreamConsumer.newFactory();
  }

  private static void destroySilently(Process p) {
    if (p == null) return;
    try {
      p.destroy();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    }
  }

  private static void waitForSilently(Process p) {
    if (p == null) return;
    boolean interrupted = false;
    try {
      for (;;) {
        try {
          p.waitFor();
          break;
        } catch (InterruptedException e) {
          interrupted = true;
        }
      }
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    } finally {
      if (interrupted) Thread.currentThread().interrupt();
    }
  }

  private static void cancelSilently(Future<?> f) {
    if (f == null) return;
    try {
      f.cancel(true);
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    }
  }

  private static void shutdownNowSilently(ExecutorService s) {
    if (s == null) return;
    try {
      s.shutdownNow();
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    }
  }

  private static void awaitTerminationSilently(ExecutorService s) {
    if (s == null) return;
    boolean interrupted = false;
    try {
      for (;;) {
        try {
          s.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
          break;
        } catch (InterruptedException e) {
          interrupted = true;
        }
      }
    } catch (ThreadDeath td) {
      throw td;
    } catch (Throwable t) {
      /* IGNORED */
    } finally {
      if (interrupted) Thread.currentThread().interrupt();
    }
  }
}
